#region Copyright Syncfusion Inc. 2001 - 2005
//
//  Copyright Syncfusion Inc. 2001 - 2005. All rights reserved.
//
//  Use of this code is subject to the terms of our license.
//  A copy of the current license can be obtained at any time by e-mailing
//  licensing@syncfusion.com. Re-distribution in any form is strictly
//  prohibited. Any infringement will be prosecuted under applicable laws. 
//
#endregion

#region file using directives
using System;
using System.Collections;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Windows.Forms;
using System.Diagnostics;
using System.Text;

using Syncfusion.ExcelRW;
using Syncfusion.ExcelRW.Implementation;
using Syncfusion.Windows.Forms.Grid;
using Syncfusion.Drawing;
using Syncfusion.Windows.Forms.Grid.Grouping;
using Syncfusion.Grouping;
#endregion

namespace Syncfusion.GridExcelConverter
{
  /// <summary>
  /// Summary description for UserControl1.
  /// </summary>
  public class GridExcelConverterControl : System.ComponentModel.Component
  {
    #region Class constants
    /// <summary>
    /// General format in Excel.
    /// </summary>
    private const int DEF_EXCEL_GENERAL_FORMAT = 0;
    /// <summary>
    /// 
    /// </summary>
    private const int DEF_EXCEL_CURRENCY_FORMAT = 5;
    /// <summary>
    /// 
    /// </summary>
    private const int DEF_EXCEL_DECIMAL_FORMAT = 4;
    /// <summary>
    /// 
    /// </summary>
    private const int DEF_EXCEL_SCIENTIFIC_FORMAT = 11;
    /// <summary>
    /// 
    /// </summary>
    private const int DEF_EXCEL_FIXEDPOINT_FORMAT = 4;
    /// <summary>
    /// 
    /// </summary>
    private const int DEF_EXCEL_NUMBER_FORMAT = 4;
    /// <summary>
    /// 
    /// </summary>
    private const int DEF_EXCEL_PERCENT_FORMAT = 9;
    /// <summary>
    /// 
    /// </summary>
    private const int DEF_EXCEL_HEX_FORMAT = 0;

    /// <summary>
    /// 
    /// </summary>
    private const char DEF_CHAR_CURRENCY = 'C';
    /// <summary>
    /// 
    /// </summary>
    private const char DEF_CHAR_DECIMAL = 'D';
    /// <summary>
    /// 
    /// </summary>
    private const char DEF_CHAR_SCIENTIFIC = 'E';
    /// <summary>
    /// 
    /// </summary>
    private const char DEF_CHAR_FIXEDPOINT = 'F';
    /// <summary>
    /// 
    /// </summary>
    private const char DEF_CHAR_GENERAL = 'G';
    /// <summary>
    /// 
    /// </summary>
    private const char DEF_CHAR_NUMBER = 'N';
    /// <summary>
    /// 
    /// </summary>
    private const char DEF_CHAR_PERCENT = 'P';
    /// <summary>
    /// 
    /// </summary>
    private const char DEF_CHAR_HEX = 'X';
    /// <summary>
    /// 
    /// </summary>
    private static readonly DisplayElementKind[] DEF_ADJUSTGROUPLEVELIFEQUAL = new DisplayElementKind[]
    {
      DisplayElementKind.Caption,
      DisplayElementKind.Table,
      DisplayElementKind.NestedTable,
    };
    #endregion

    #region Class members
    //    /// <summary>
    //    /// 
    //    /// </summary>
    //    private static readonly Hashtable s_hashBorderIndexToExcelLineStyle = new Hashtable();
    #endregion

    #region Class Initialize/Finalize methods
    /// <summary>
    /// Static constructor.
    /// </summary>
    static GridExcelConverterControl()
    {
      FormulaUtil.RegisterAdditionalAlias( "POW", ExcelFunction.POWER );
      FormulaUtil.RegisterAdditionalAlias( "AVG", ExcelFunction.AVERAGE );
    }
    //    /// <summary>
    //    /// Static constructor.
    //    /// </summary>
    //    static GridExcelConverterControl()
    //    {
    //      FillBordersHash();
    //    }
    /// <summary>
    /// Default constructor.
    /// </summary>
    public GridExcelConverterControl()
    {
    }

    /// <summary>
    /// Clean up any resources being used.
    /// </summary>
    protected override void Dispose( bool disposing )
    {
      base.Dispose( disposing );
    }

    #endregion

    #region Class Export Grid to Excel methods
    /// <summary>
    /// Converts grid into excel file.
    /// </summary>
    /// <param name="grid">Grid to export.</param>
    /// <param name="fileName">File to export into.</param>
    public void GridToExcel( GridControl grid, string fileName )
    {
      GridToExcel( grid.Model, fileName );
    }
    /// <summary>
    /// Converts grid into excel file.
    /// </summary>
    /// <param name="grid">Grid to export.</param>
    /// <param name="sheet">Destination worksheet.</param>
    public void GridToExcel( GridControl grid, IWorksheet sheet )
    {
      GridToExcel( grid.Model, sheet );
    }
    /// <summary>
    /// Converts grid into excel file.
    /// </summary>
    /// <param name="grid">Grid to export.</param>
    /// <param name="fileName">File to export into.</param>
    public void GridToExcel( GridModel grid, string fileName )
    {
      ExcelEngine engine = new ExcelEngine();
      IWorkbook book = engine.Excel.Workbooks.Create( 1 );
      IWorksheet sheet = book.Worksheets[ 0 ];

      GridToExcel( grid, sheet );

      book.Close( true, fileName );
    }
    /// <summary>
    /// Converts grid into excel worksheet.
    /// </summary>
    /// <param name="grid">Grid to export.</param>
    /// <param name="sheet">Destination worksheet.</param>
    public void GridToExcel( GridModel grid, IWorksheet sheet )
    {
      if( grid == null )
        throw new ArgumentNullException( "grid" );

      if( sheet == null )
        throw new ArgumentNullException( "sheet" );

      CopyMergesFromGrid( grid, sheet );
      CopyRowHeightFromGrid( grid, sheet );
      CopyColumnWidthFromGrid( grid, sheet );

      for( int iRow = 1, iRowCount = grid.RowCount; iRow <= iRowCount; iRow++ )
      {
        for( int iColumn = 1, iColCount = grid.ColCount; iColumn <= iColCount; iColumn++ )
        {
          GridStyleInfo cell = grid[ iRow, iColumn ];
          GridCellToExcel( grid, cell, sheet.Range[ iRow, iColumn ] );
        }
      }
    }
    /// <summary>
    /// Copies merged ranges.
    /// </summary>
    /// <param name="grid">Source grid.</param>
    /// <param name="sheet">Destination worksheet.</param>
    private void CopyMergesFromGrid( GridModel grid, IWorksheet sheet )
    {
      if( grid == null )
        throw new ArgumentNullException( "grid" );

      if( sheet == null )
        throw new ArgumentNullException( "sheet" );

      GridModelCoveredRanges coveredRanges = grid.CoveredRanges;
      GridRangeInfoList ranges = coveredRanges.Ranges;

      for( int i = 0, len = ranges.Count; i < len; i++ )
      {
        GridRangeInfo rangeInfo = ranges[ i ];
        Debug.WriteLine( i, "Range index" );

        string strMessage = string.Format( "({0}, {1})-({2}, {3})", rangeInfo.Top, rangeInfo.Left, rangeInfo.Bottom, rangeInfo.Right );
        Debug.WriteLine( strMessage, "Range rect" );

        if( rangeInfo.Top == 0 || rangeInfo.Left == 0 ) continue;

        sheet.Range[ rangeInfo.Top, rangeInfo.Left, rangeInfo.Bottom, rangeInfo.Right ].Merge();
      }
    }
    /// <summary>
    /// Copies one grid cell into excel cell.
    /// </summary>
    /// <param name="grid">Source grid.</param>
    /// <param name="gridCell">Grid cell to copy.</param>
    /// <param name="range">Destination excel range.</param>
    private void GridCellToExcel( GridModel grid, GridStyleInfo gridCell, IRange range )
    {
      if( grid == null )
        throw new ArgumentNullException( "grid" );

      if( gridCell == null )
        throw new ArgumentNullException( "gridCell" );

      if( range == null )
        throw new ArgumentNullException( "range" );

      range.Value = gridCell.Text;
      CopyStyles( grid, gridCell, range );
    }
    /// <summary>
    /// Copies style of grid cell into excel cell.
    /// </summary>
    /// <param name="grid">Source grid.</param>
    /// <param name="gridCell">Source cell to copy.</param>
    /// <param name="range">Destination range.</param>
    private void CopyStyles( GridModel grid, GridStyleInfo gridCell, IRange range )
    {
      if( gridCell == null )
        throw new ArgumentNullException( "gridCell" );

      if( range == null )
        throw new ArgumentNullException( "range" );

      GridStyleInfo[] arrStyles = new GridStyleInfo[ 8 ]
      {
        gridCell,
        null,
        grid.RowStyles[ range.Row ],
        null,
        grid.ColStyles[ range.Column ],
        null,
        grid.TableStyle,
        null,
      };

      int iStylesCount = arrStyles.Length;
      for( int i = 0; i < iStylesCount; i+= 2 )
      {
        GridStyleInfo style = arrStyles[ i ];

        if( style.HasBaseStyle )
        {
          arrStyles[ i + 1 ] = grid.BaseStylesMap[ style.BaseStyle ].StyleInfo;
        }
      }

      GridStyleInfo finalGridStyle = new GridStyleInfo( arrStyles[ 0 ] );

      for( int i = 1; i < iStylesCount; i++ )
      {
        GridStyleInfo style = arrStyles[ i ];

        if( style == null ) continue;

        if( !CombineStyles( finalGridStyle, style ) ) break;
      }

      CopyStyle( finalGridStyle, range );
    }
    /// <summary>
    /// Copies style from grid style into range style.
    /// </summary>
    /// <param name="gridStyle">Grid style to copy.</param>
    /// <param name="destRange">Destination range.</param>
    private void CopyStyle( GridStyleInfo gridStyle, IRange destRange )
    {
      if( gridStyle == null )
        throw new ArgumentNullException( "gridStyle" );

      if( destRange == null )
        throw new ArgumentNullException( "destRange" );

      CopyBrush( gridStyle, destRange );
      CopyFont( gridStyle, destRange );
      CopyBorders( gridStyle, destRange );
      CopyAlignment( gridStyle, destRange );

      if( gridStyle.HasFormat )
      {
        destRange.CellStyle.NumberFormatIndex = ConvertFormatFromGridToExcel( gridStyle.Format );
      }
    }
    /// <summary>
    /// Copies brush settings from grid cell into excel cell.
    /// </summary>
    /// <param name="gridCell">Source grid cell.</param>
    /// <param name="range">Destination excel cell.</param>
    private void CopyBrush( GridStyleInfo gridCell, IRange range )
    {
      // TODO: finish implementation.
      if( gridCell.Interior.Style == BrushStyle.Pattern )
      {
        IStyle rangeStyle = range.CellStyle;
        rangeStyle.FillPattern = GetClosestExcelPattern( gridCell.Interior.PatternStyle );
        rangeStyle.FillForegroundRGB = gridCell.Interior.ForeColor;
        rangeStyle.FillBackgroundRGB = gridCell.Interior.BackColor;
      }
      else if( gridCell.Interior.Style == BrushStyle.Solid )
      {
        IStyle rangeStyle = range.CellStyle;
        rangeStyle.FillForegroundRGB = gridCell.Interior.ForeColor;
        rangeStyle.FillBackgroundRGB = gridCell.Interior.BackColor;
        Debug.WriteLine( rangeStyle.FillBackground, "FillBackgroud" );
      }

      //      if( !gridCell.Interior.BackColor.IsEmpty )
      //      {
      //        xStyle.FillBackground = myWorkBook.SetColorOrGetNearest(style.BackColor);
      //      }
    }
    /// <summary>
    /// Copies font settings from grid cell into excel cell.
    /// </summary>
    /// <param name="gridCell">Source grid cell.</param>
    /// <param name="range">Destination excel cell.</param>
    private void CopyFont( GridStyleInfo gridCell, IRange range )
    {
      if( gridCell.HasFont )
      {
        IStyle rangeStyle = range.CellStyle;
        IFont rangeFont = rangeStyle.Font;
        GridFontInfo font = gridCell.Font;

        rangeFont.RGBColor = gridCell.TextColor;
        rangeFont.Size = Convert.ToInt32( font.Size );
            
        if( font.HasFacename )
        {
          rangeFont.FontName = font.Facename;
        }

        if( font.HasStrikeout )
        {
          rangeFont.Strikethrough = font.Strikeout;
        }

        if( font.HasUnderline )
        {
          rangeFont.Underline = ExcelUnderline.Single;
        }

        if( font.HasItalic )
        {
          rangeFont.Italic = font.Italic;
        }

        if( font.HasBold )
        {
          rangeFont.Bold = font.Bold;
        }
      }
    }
    /// <summary>
    /// Copies borders settings from grid cell into excel cell.
    /// </summary>
    /// <param name="gridCell">Source grid cell.</param>
    /// <param name="range">Destination excel cell.</param>
    private void CopyBorders( GridStyleInfo gridCell, IRange range )
    {
      if( gridCell == null )
        throw new ArgumentNullException( "gridCell" );

      if( range == null )
        throw new ArgumentNullException( "range" );

      if( gridCell.HasBorders )
      {
        GridBordersInfo borders = gridCell.Borders;
        IBorders rangeBorders = range.Borders;

        if( borders.HasBottom )
        {
          CopyBorder( borders.Bottom, rangeBorders[ ExcelBordersIndex.EdgeBottom ] );
        }

        if( borders.HasLeft )
        {
          CopyBorder( borders.Left, rangeBorders[ ExcelBordersIndex.EdgeLeft ] );
        }

        if( borders.HasRight )
        {
          CopyBorder( borders.Right, rangeBorders[ ExcelBordersIndex.EdgeRight ] );
        }

        if( borders.HasTop )
        {
          CopyBorder( borders.Top, rangeBorders[ ExcelBordersIndex.EdgeTop ] );
        }
      }
    }
    /// <summary>
    /// Copies alignment settings from grid cell into excel cell..
    /// </summary>
    /// <param name="gridCell">Source grid cell.</param>
    /// <param name="range">Destination excel cell.</param>
    private void CopyAlignment( GridStyleInfo gridCell, IRange range )
    {
      IStyle rangeStyle = range.CellStyle;

      if( gridCell.HasHorizontalAlignment )
      {
        rangeStyle.HorizontalAlignment = GetClosestExcelHorizontalAlignment( gridCell.HorizontalAlignment );
      }

      if( gridCell.HasVerticalAlignment )
      {
        rangeStyle.VerticalAlignment = GetClosestExcelVerticalAlignment( gridCell.VerticalAlignment );
      }

      if( gridCell.HasWrapText )
      { 
        rangeStyle.WrapText = gridCell.WrapText;
      }
    }
    /// <summary>
    /// Copies border settings from grid border into excel border.
    /// </summary>
    /// <param name="cellBorder">Grid border to copy.</param>
    /// <param name="rangeBorder">Destination excel border.</param>
    private void CopyBorder( GridBorder cellBorder, IBorder rangeBorder )
    {
      if( cellBorder == null )
        throw new ArgumentNullException( "cellBorder" );

      if( rangeBorder == null )
        throw new ArgumentNullException( "rangeBorder" );

      rangeBorder.ColorRGB = cellBorder.Color;
      //int iBorderTypeIndex = GetIndex( cellBorder );
      //ExcelLineStyle lineStyle = s_hashBorderIndexToExcelLineStyle[ iBorderTypeIndex ];
      rangeBorder.LineStyle = GetExcelLineStyle( cellBorder.Style );
    }
    /// <summary>
    /// Converts GridBorderStyle into ExcelLineStyle.
    /// </summary>
    /// <param name="gridLineStyle">GridBorderStyle to convert.</param>
    /// <returns>Corresponding ExcelLineStyle.</returns>
    private ExcelLineStyle GetExcelLineStyle( GridBorderStyle gridLineStyle )
    {
      switch ( gridLineStyle )
      {
        case GridBorderStyle.DashDot:    return ExcelLineStyle.Dash_dot;
        case GridBorderStyle.DashDotDot: return ExcelLineStyle.Dash_dot_dot;
        case GridBorderStyle.Dashed:     return ExcelLineStyle.Dashed;
        case GridBorderStyle.Dotted:     return ExcelLineStyle.Dotted;
        case GridBorderStyle.None:       return ExcelLineStyle.None;
        case GridBorderStyle.NotSet:     return ExcelLineStyle.None;
        case GridBorderStyle.Solid:      return ExcelLineStyle.Thick;
        case GridBorderStyle.Standard:   return ExcelLineStyle.Thin;

        default: throw new ArgumentOutOfRangeException( "Border style is out of range" );
      }
    }

    /// <summary>
    /// Converts PatterStyle into ExcelPattern.
    /// </summary>
    /// <param name="gridPattern">PatternStyle to convert.</param>
    /// <returns>Corresponding ExcelPattern.</returns>
    private ExcelPattern GetClosestExcelPattern( PatternStyle gridPattern )
    {
      switch( gridPattern )
      {
        case PatternStyle.BackwardDiagonal:     return ExcelPattern.LightUpwardDiagonal;
        case PatternStyle.Cross:                return ExcelPattern.Angle;
        case PatternStyle.DarkDownwardDiagonal: return ExcelPattern.DarkDownwardDiagonal;
        case PatternStyle.DarkHorizontal:       return ExcelPattern.DarkHorizontal;
        case PatternStyle.DarkUpwardDiagonal:   return ExcelPattern.DarkUpwardDiagonal;
        case PatternStyle.DashedVertical:       return ExcelPattern.Vertical;
        case PatternStyle.DiagonalBrick:        return ExcelPattern.LightUpwardDiagonal;
        case PatternStyle.DiagonalCross:        return ExcelPattern.Percent60;
        case PatternStyle.Divot:                return ExcelPattern.Percent05;
        case PatternStyle.DottedDiamond:        return ExcelPattern.Percent10;
        case PatternStyle.DottedGrid:           return ExcelPattern.Angle;
        case PatternStyle.ForwardDiagonal:      return ExcelPattern.ForwardDiagonal;
        case PatternStyle.Horizontal:           return ExcelPattern.Horizontal;
        case PatternStyle.HorizontalBrick:      return ExcelPattern.Angle;
        case PatternStyle.LargeCheckerBoard:    return ExcelPattern.Angle;
        case PatternStyle.LargeConfetti:        return ExcelPattern.Percent05;
        case PatternStyle.LightDownwardDiagonal:return ExcelPattern.LightDownwardDiagonal;
        case PatternStyle.LightHorizontal:      return ExcelPattern.Horizontal;
        case PatternStyle.LightUpwardDiagonal:  return ExcelPattern.LightUpwardDiagonal;
        case PatternStyle.LightVertical:        return ExcelPattern.Vertical;
        case PatternStyle.NarrowHorizontal:     return ExcelPattern.Horizontal;
        case PatternStyle.NarrowVertical:       return ExcelPattern.Vertical;
        case PatternStyle.None:                 return ExcelPattern.None;
        case PatternStyle.OutlinedDiamond:      return ExcelPattern.Percent75;
        case PatternStyle.Percent05:            return ExcelPattern.Percent05;
        case PatternStyle.Percent10:            return ExcelPattern.Percent10;
        case PatternStyle.Percent20:            return ExcelPattern.Percent25;
        case PatternStyle.Percent25:            return ExcelPattern.Percent25;
        case PatternStyle.Percent30:            return ExcelPattern.Percent25;
        case PatternStyle.Percent40:            return ExcelPattern.Percent50;
        case PatternStyle.Percent50:            return ExcelPattern.Percent50;
        case PatternStyle.Percent60:            return ExcelPattern.Percent60;
        case PatternStyle.Percent70:            return ExcelPattern.Percent70;
        case PatternStyle.Percent75:            return ExcelPattern.Percent75;
        case PatternStyle.Percent80:            return ExcelPattern.Percent75;
        case PatternStyle.Percent90:            return ExcelPattern.Percent75;
        case PatternStyle.Plaid:                return ExcelPattern.Angle;
        case PatternStyle.Shingle:              return ExcelPattern.Angle;
        case PatternStyle.SmallCheckerBoard:    return ExcelPattern.ForwardDiagonal;
        case PatternStyle.SmallConfetti:        return ExcelPattern.Percent10;
        case PatternStyle.SmallGrid:            return ExcelPattern.Angle;
        case PatternStyle.SolidDiamond:         return ExcelPattern.Percent75;
        case PatternStyle.Sphere:               return ExcelPattern.ForwardDiagonal;
        case PatternStyle.Trellis:              return ExcelPattern.Percent75;
        case PatternStyle.Vertical:             return ExcelPattern.Vertical;
        case PatternStyle.Wave:                 return ExcelPattern.ForwardDiagonal;
        case PatternStyle.Weave:                return ExcelPattern.ForwardDiagonal;
        case PatternStyle.WideDownwardDiagonal: return ExcelPattern.LightDownwardDiagonal;
        case PatternStyle.WideUpwardDiagonal:   return ExcelPattern.LightUpwardDiagonal;
        case PatternStyle.ZigZag:               return ExcelPattern.Horizontal;

        default: throw new ArgumentOutOfRangeException( "PatternStyle" );
      }
    }

//    /// <summary>
//    /// 
//    /// </summary>
//    /// <param name="cellBorder"></param>
//    /// <returns></returns>
//    private int GetIndex( GridBorder cellBorder )
//    {
//      if( cellBorder == null )
//        throw new ArgumentNullException( "cellBorder" );
//
//      ushort usWeight = ( ushort )cellBorder.Weight;
//      ushort usLineStyle = ( ushort )cellBorder.Style;
//      int iResult = usWeight << 16 + usLineStyle;
//
//      return iResult;
//    }
    /// <summary>
    /// Converts GridHorizontalAlignment into ExcelHAlign.
    /// </summary>
    /// <param name="gridHAlign">GridHorizontalAlignment to convert.</param>
    /// <returns>Corresponding ExcelHAlign.</returns>
    private ExcelHAlign GetClosestExcelHorizontalAlignment( GridHorizontalAlignment gridHAlign )
    {
      switch( gridHAlign )
      {
        case GridHorizontalAlignment.Center: return ExcelHAlign.HAlignCenter;
        case GridHorizontalAlignment.Left:   return ExcelHAlign.HAlignLeft;
        case GridHorizontalAlignment.Right:  return ExcelHAlign.HAlignRight;

        default: throw new ArgumentOutOfRangeException( "gridHAling" );
      }
    }
    /// <summary>
    /// Converts GridVerticalAlignment into ExcelVAlign.
    /// </summary>
    /// <param name="gridVAlign">GridVerticalAlignment to convert.</param>
    /// <returns>Corresponding ExcelVAlign.</returns>
    private ExcelVAlign GetClosestExcelVerticalAlignment( GridVerticalAlignment gridVAlign )
    {
      switch( gridVAlign )
      {
        case GridVerticalAlignment.Bottom: return ExcelVAlign.VAlignBottom;
        case GridVerticalAlignment.Middle: return ExcelVAlign.VAlignCenter;
        case GridVerticalAlignment.Top:    return ExcelVAlign.VAlignTop;

        default: throw new ArgumentOutOfRangeException( "gridVAlign" );
      }
    }
    /// <summary>
    /// Combines two styles.
    /// </summary>
    /// <param name="destinationStyle">Style which is primary and in which result will be accumulated.</param>
    /// <param name="baseStyle">Style to combine.</param>
    /// <returns>Value indicating that not all properties were set.</returns>
    private bool CombineStyles( GridStyleInfo destinationStyle, GridStyleInfo baseStyle )
    {
      if( destinationStyle == null )
        throw new ArgumentNullException( "destinationStyle" );

      if( baseStyle == null )
        throw new ArgumentNullException( "baseStyle" );

      bool bResult = false;

      bResult = CombineBorders( destinationStyle, baseStyle );
      bResult |= CombineFont( destinationStyle, baseStyle );
      bResult |= CombineBrush( destinationStyle, baseStyle );
      bResult |= CombineAlignment( destinationStyle, baseStyle );

      return bResult;
    }
    /// <summary>
    /// Combine borders information if necessary.
    /// </summary>
    /// <param name="destBorders">Destination style.</param>
    /// <param name="baseBorders">Source style.</param>
    /// <returns>True if not all borders in destBorders were set.</returns>
    private bool CombineBorders( GridStyleInfo destStyle, GridStyleInfo baseStyle )
    {
      if( destStyle == null )
        throw new ArgumentNullException( "destBorders" );

      if( baseStyle == null )
        throw new ArgumentNullException( "baseBorders" );

      if( !baseStyle.HasBorders ) return true;

      GridBordersInfo destBorders = destStyle.Borders;
      GridBordersInfo baseBorders = baseStyle.Borders;

      bool bResult = !destBorders.HasBottom && !baseBorders.HasBottom;

      if( !destBorders.HasBottom && baseBorders.HasBottom )
      {
        GridBorder border = baseBorders.Bottom;
        destBorders.Bottom = new GridBorder( border.Style, border.Color, border.Weight );
      }

      bResult |= !destBorders.HasLeft && !baseBorders.HasLeft;
      if( !destBorders.HasLeft && baseBorders.HasLeft )
      {
        GridBorder border = baseBorders.Left;
        destBorders.Left = new GridBorder( border.Style, border.Color, border.Weight );
      }

      bResult |= !destBorders.HasRight && !baseBorders.HasRight;
      if( !destBorders.HasRight && baseBorders.HasRight )
      {
        GridBorder border = baseBorders.Right;
        destBorders.Right = new GridBorder( border.Style, border.Color, border.Weight );
      }

      bResult |= !destBorders.HasTop && !baseBorders.HasTop;
      if( !destBorders.HasTop && baseBorders.HasTop )
      {
        GridBorder border = baseBorders.Top;
        destBorders.Top = new GridBorder( border.Style, border.Color, border.Weight );
      }

      return bResult;
    }
    /// <summary>
    /// Combine fonts if necessary.
    /// </summary>
    /// <param name="destFont">Destination style.</param>
    /// <param name="baseFont">Source style.</param>
    /// <returns>True if not all font properties in destFont were set.</returns>
    private bool CombineFont( GridStyleInfo destStyle, GridStyleInfo baseStyle )
    {
      if( destStyle == null )
        throw new ArgumentNullException( "destStyle" );

      if( baseStyle == null )
        throw new ArgumentNullException( "baseStyle" );

      bool bResult = false;

      if( !destStyle.HasTextColor )
      {
        if( baseStyle.HasTextColor )
        {
          destStyle.TextColor = baseStyle.TextColor;
        }
        else
        {
          bResult = true;
        }
      }

      if( !baseStyle.HasFont ) return true;

      GridFontInfo destFont = destStyle.Font;
      GridFontInfo baseFont = baseStyle.Font;

      if( !destFont.HasBold )
      {
        if( baseFont.HasBold )
        {
          destFont.Bold = baseFont.Bold;
        }
        else
        {
          bResult = true;
        }
      }

      if( !destFont.HasItalic )
      {
        if( baseFont.HasItalic )
        {
          destFont.Italic = baseFont.Italic;
        }
        else
        {
          bResult = true;
        }
      }

      if( !destFont.HasUnderline )
      {
        if( baseFont.HasUnderline )
        {
          destFont.Underline = baseFont.Underline;
        }
        else
        {
          bResult = true;
        }
      }

      if( !destFont.HasFacename )
      {
        if( baseFont.HasFacename )
        {
          destFont.Facename = baseFont.Facename;
        }
        else
        {
          bResult = true;
        }
      }

      if( !destFont.HasOrientation )
      {
        if( baseFont.HasOrientation )
        {
          destFont.Orientation = baseFont.Orientation;
        }
        else
        {
          bResult = true;
        }
      }

      if( !destFont.HasSize )
      {
        if( baseFont.HasSize )
        {
          destFont.Size = baseFont.Size;
        }
        else
        {
          bResult = true;
        }
      }

      if( !destFont.HasStrikeout )
      {
        if( baseFont.HasStrikeout )
        {
          destFont.Strikeout = baseFont.Strikeout;
        }
        else
        {
          bResult = true;
        }
      }

      if( !destFont.HasUnit )
      {
        if( baseFont.HasUnit )
        {
          destFont.Unit = baseFont.Unit;
        }
        else
        {
          bResult = true;
        }
      }

      return bResult;
    }
    /// <summary>
    /// Combines brushes if necessary.
    /// </summary>
    /// <param name="destStyle">Destination style.</param>
    /// <param name="baseStyle">Source style.</param>
    /// <returns>True if not all font properties in destFont were set.</returns>
    private bool CombineBrush( GridStyleInfo destStyle, GridStyleInfo baseStyle )
    {
      if( destStyle == null )
        throw new ArgumentNullException( "destStyle" );

      if( baseStyle == null )
        throw new ArgumentNullException( "baseStyle" );

      if( !baseStyle.HasInterior ) return true;

      if( destStyle.HasInterior ) return false;

      destStyle.Interior = new BrushInfo( baseStyle.Interior );

      return false;
    }
    /// <summary>
    /// Combines alignment options if necessary.
    /// </summary>
    /// <param name="destStyle">Destination style.</param>
    /// <param name="baseStyle">Source style.</param>
    /// <returns>True if not all font properties in destFont were set.</returns>
    private bool CombineAlignment( GridStyleInfo destStyle, GridStyleInfo baseStyle )
    {
      if( destStyle == null )
        throw new ArgumentNullException( "destStyle" );

      if( baseStyle == null )
        throw new ArgumentNullException( "baseStyle" );

      bool bResult = false;

      bResult |= !destStyle.HasHorizontalAlignment && !baseStyle.HasHorizontalAlignment;
      if( !destStyle.HasHorizontalAlignment && baseStyle.HasHorizontalAlignment )
      {
        destStyle.HorizontalAlignment = baseStyle.HorizontalAlignment;
      }

      bResult |= !destStyle.HasVerticalAlignment && !baseStyle.HasVerticalAlignment;
      if( !destStyle.HasVerticalAlignment && baseStyle.HasVerticalAlignment )
      {
        destStyle.VerticalAlignment = baseStyle.VerticalAlignment;
      }

      bResult |= !destStyle.HasWrapText && !baseStyle.HasWrapText;
      if( !destStyle.HasWrapText && baseStyle.HasWrapText )
      { 
        destStyle.WrapText = baseStyle.WrapText;
      }

      return bResult;
    }
    /// <summary>
    /// Copies row height settings from grid model into excel worksheet.
    /// </summary>
    /// <param name="grid">Source grid.</param>
    /// <param name="sheet">Destination worksheet.</param>
    private void CopyRowHeightFromGrid( GridModel grid, IWorksheet sheet )
    {
      if( grid == null )
        throw new ArgumentNullException( "grid" );

      if( sheet == null )
        throw new ArgumentNullException( "sheet" );

      for( int i = 1, len = grid.RowCount; i < len; i++ )
      {
        int iRowHeight = grid.RowHeights[ i ];
        sheet.SetRowHeightInPixels( i, iRowHeight );
      }
    }

    /// <summary>
    /// Copies column width settings from grid model into excel worksheet.
    /// </summary>
    /// <param name="grid">Source grid.</param>
    /// <param name="sheet">Destination worksheet.</param>
    private void CopyColumnWidthFromGrid( GridModel grid, IWorksheet sheet )
    {
      CopyColumnWidthFromGrid( grid, sheet, 1 );
    }
    /// <summary>
    /// Copies column width settings from grid model into excel worksheet.
    /// </summary>
    /// <param name="grid">Source grid.</param>
    /// <param name="sheet">Destination worksheet.</param>
    /// <param name="iStartGridColumn">Start column in the grid.</param>
    private void CopyColumnWidthFromGrid( GridModel grid, IWorksheet sheet, int iStartGridColumn )
    {
      if( grid == null )
        throw new ArgumentNullException( "grid" );

      if( sheet == null )
        throw new ArgumentNullException( "sheet" );

      for( int i = iStartGridColumn, len = grid.ColCount; i < len; i++ )
      {
        int iWidthInPixels = grid.Model.ColWidths[ i ];
        double dWidth = sheet.PixelsToColumnWidth( iWidthInPixels );
        sheet.SetColumnWidth( i - iStartGridColumn + 1, dWidth );
      }
    }
    /// <summary>
    /// Converts grid number format into excel number format.
    /// </summary>
    /// <param name="strGridFormat">Grid format to convert.</param>
    /// <returns>Index to the excel format.</returns>
    private int ConvertFormatFromGridToExcel( string strGridFormat )
    {
      if( strGridFormat == null || strGridFormat.Length == 0 )
        return DEF_EXCEL_GENERAL_FORMAT;

      // TODO: implement better format converter.

      switch( char.ToUpper( strGridFormat[ 0 ] ) )
      {
        case DEF_CHAR_GENERAL:
          return DEF_EXCEL_GENERAL_FORMAT;

        case DEF_CHAR_CURRENCY:
          return DEF_EXCEL_CURRENCY_FORMAT;

        case DEF_CHAR_DECIMAL:
          return DEF_EXCEL_DECIMAL_FORMAT;

        case DEF_CHAR_SCIENTIFIC:
          return DEF_EXCEL_SCIENTIFIC_FORMAT;

        case DEF_CHAR_FIXEDPOINT:
          return DEF_EXCEL_FIXEDPOINT_FORMAT;

        case DEF_CHAR_NUMBER:
          return DEF_EXCEL_NUMBER_FORMAT;

        case DEF_CHAR_PERCENT:
          return DEF_EXCEL_PERCENT_FORMAT;

        case DEF_CHAR_HEX:
          return DEF_EXCEL_HEX_FORMAT;

        default:
          return DEF_EXCEL_GENERAL_FORMAT;
      }
    }
    #endregion

    #region Export Grouping Grid to Excel methods
    /// <summary>
    /// Converts grouping grid into excel file.
    /// </summary>
    /// <param name="grouping">Source grouping grid.</param>
    /// <param name="strFileName">Destination worksheet.</param>
    /// <param name="options"></param>
    public void GroupingGridToExcel( GridGroupingControl grouping, string strFileName,
      ConverterOptions options )
    {
      if( grouping == null )
        throw new ArgumentNullException( "grouping" );

      if( strFileName == null )
        throw new ArgumentNullException( "strFileName" );

      if( strFileName.Length == 0 )
        throw new ArgumentException( "strFileName - string can not be empty" );

      ExcelEngine engine = new ExcelEngine();
      IWorkbook book = engine.Excel.Workbooks.Create( 1 );
      IWorksheet sheet = book.Worksheets[ 0 ];

      GroupingGridToExcel( grouping, sheet, options );

      book.Close( true, strFileName );
    }
    /// <summary>
    /// Converts grouping grid into excel file.
    /// </summary>
    /// <param name="grouping">Source grouping grid.</param>
    /// <param name="sheet">Destination worksheet.</param>
    /// <param name="options"></param>
    public void GroupingGridToExcel( GridGroupingControl grouping, IWorksheet sheet,
      ConverterOptions options )
    {
      if( grouping == null )
        throw new ArgumentNullException( "grouping" );

      if( sheet == null )
        throw new ArgumentNullException( "sheet" );

      sheet.PageSetup.IsSummaryRowBelow = false;
      GridTable table = grouping.Table;

      if( options == ConverterOptions.ConvertVisible )
      {
        ExportElements( table.DisplayElements, sheet, 0, options );
      }
      else
      {
        ExportElements( table.Elements, sheet, 0, options );
      }

      GridTableModel tableModel = grouping.TableModel;
      CopyRowHeightFromGrid( tableModel, sheet );
      CopyColumnWidthFromGrid( tableModel, sheet, tableModel.GetColumnIndentCount() );

//      GridColumnHeaderRow header;
//      header.ToString();
//
//      FieldDescriptorCollection fields = header.ParentTable.TableDescriptor.Fields;
    }
    /// <summary>
    /// Exports collection of elements into worksheet.
    /// </summary>
    /// <param name="arrElements">Collection of elements to export.</param>
    /// <param name="sheet">Destination worksheet.</param>
    /// <param name="index">Starting zero-based index to the first row.</param>
    /// <param name="options">Convert options.</param>
    /// <returns>Index to the row after last exported element.</returns>
    private int ExportElements( IList arrElements, IWorksheet sheet, int index, ConverterOptions options )
    {
      if( arrElements == null )
        throw new ArgumentNullException( "arrElements" );

      if( sheet == null )
        throw new ArgumentNullException( "worksheet" );

      if( index < 0 )
        throw new ArgumentOutOfRangeException( "index", index, "Value can not be less 0" );

      if( arrElements.Count == 0 ) return 0;

      Element element = ( Element )arrElements[ 0 ];
      GridTableDescriptor tableDesc = ( GridTableDescriptor )element.ParentTable.TableDescriptor;
      int iColumnCount = tableDesc.Columns.Count;

      //for( int i = 0, len = arrElements.Count; i < len; i++ )
      int iIndexInGroup = 0;
      int i = 0;
      int iCount = arrElements.Count;

      Stack stackStartIndexes = new Stack();
      int iCurLevel = -1;

      while( i < iCount )
      {
        element = ( Element )arrElements[ i ];
        int iElementLevel = element.GroupLevel;

        if( iElementLevel != iCurLevel &&
          Array.IndexOf( DEF_ADJUSTGROUPLEVELIFEQUAL, element.Kind ) == -1 )
        {
          AdjustGroupingLevel( stackStartIndexes, ref iCurLevel, iElementLevel, sheet, index + i, false );
        }

        switch( element.Kind )
        {
          case DisplayElementKind.Table:
            GridTable table = ( GridTable )element;

            IList elements = null;

            if( options == ConverterOptions.ConvertVisible )
            {
              elements = table.DisplayElements;
            }
            else
            {
              elements = table.Elements;
            }

            i += ExportElements( elements, sheet, index + i, options );
            break;

          case DisplayElementKind.Empty:
            GridEmptySection empty = ( GridEmptySection )element;
            i += ExportEmptyCells( sheet, index + i, options,
              empty.Appearance.EmptyCell, iColumnCount );
            break;

          case DisplayElementKind.AddNewRecord:
            RecordRow record = ( RecordRow )element;
            i += ExportRecordRow( record, sheet, index + i, options, iIndexInGroup );
            break;

          case DisplayElementKind.Record:
            i += ExportRecordRow( ( RecordRow )element, sheet, index + i, options, iIndexInGroup );
            iIndexInGroup++;
            break;

          case DisplayElementKind.Caption:
            iIndexInGroup = 0;
            GridCaptionRow captionRow = ( GridCaptionRow )element;
            i += ExportCaption( captionRow, sheet, index + i, options, ref iCurLevel, stackStartIndexes );
            break;

          case DisplayElementKind.Summary:
            GridSummaryRow summary = ( GridSummaryRow )element;
            i += ExportSummaryRow( summary, sheet, index + i, options );
            break;

          case DisplayElementKind.NestedTable:
            NestedTable nestedTable = ( NestedTable )element;
            i += ExportNestedTable( nestedTable, sheet, index + i, options );
            break;

          case DisplayElementKind.None:
          case DisplayElementKind.FilterBar:
          case DisplayElementKind.ColumnHeader:
          case DisplayElementKind.GroupHeader:
          case DisplayElementKind.GroupFooter:
          case DisplayElementKind.RecordPreview:
          case DisplayElementKind.GroupPreview:
            sheet.Range[ index + i + 1, 1 ].Text = element.Kind.ToString();
            sheet.Range[ index + i + 1, 2 ].Text = element.Info;
            i++;
            break;
        }
      }

      if( iCurLevel >= 0 )
      {
        AdjustGroupingLevel( stackStartIndexes, ref iCurLevel, 0, sheet, index + i, false );
      }

      //sheet.Range[ index + 1, 1, index + iCount, 1 ].Group( ExcelGroupBy.ByRows, false );
      return iCount;
    }

    /// <summary>
    /// 
    /// </summary>
    /// <param name="empty"></param>
    /// <param name="sheet"></param>
    /// <param name="index"></param>
    /// <param name="options"></param>
    /// <returns></returns>
    private int ExportEmptyCells( IWorksheet sheet, int index,
      ConverterOptions options, GridTableCellStyleInfo style, int iColumnCount )
    {
      if( sheet == null )
        throw new ArgumentNullException( "sheet" );

      if( index < 0 )
        throw new ArgumentOutOfRangeException( "index", index, "Value can not be less 0" );

      IRange usedRange = sheet.Range;
      int iRowIndex = index + 1;
      IRange range = sheet.Range[ iRowIndex, 1, iRowIndex, iColumnCount ];
      CopyStyle( style, range );
      return 1;
    }
    /// <summary>
    /// Exports summary row into excel worksheet.
    /// </summary>
    /// <param name="summary">Summary to export.</param>
    /// <param name="sheet">Destination worksheet.</param>
    /// <param name="index">Zero-based index to first row of the summary.</param>
    /// <param name="options">Converter options.</param>
    private int ExportSummaryRow( GridSummaryRow summary, IWorksheet sheet, int index,
      ConverterOptions options )
    {
      if( summary == null )
        throw new ArgumentNullException( "summary" );

      if( sheet == null )
        throw new ArgumentNullException( "sheet" );

      if( index < 0 )
        throw new ArgumentOutOfRangeException( "index", index, "Value can not be less 0" );

      int iRow = index + 1;
      GridSummaryRowDescriptor descriptor = summary.SummaryRowDescriptor;
      IRange range = sheet.Range[ iRow, 1 ];
      range.Text = descriptor.Title;
      CopyStyle( descriptor.Appearance.SummaryTitleCell, range );

      //GridSummaryColumnDescriptorCollection arrColumns = descriptor.SummaryColumns;
      GridTableDescriptor table = summary.ParentTableDescriptor;
      GridSummaryColumnDescriptorCollection arrColumns = descriptor.SummaryColumns;

      int iColumnsCount = table.Columns.Count;
      range = sheet.Range[ iRow, 2, iRow, iColumnsCount ];
      CopyStyle( descriptor.Appearance.SummaryEmptyCell, range );

      for( int i = 0, len = arrColumns.Count; i < len; i++ )
      {
        GridSummaryColumnDescriptor column = arrColumns[ i ];
        int iColumnIndex = column.ColInRecord;
        GridTableCellStyleInfo style = null;

        if( iColumnIndex == -1 )
        {
          range = sheet.Range[ iRow, 1 ];
        }
        else
        {
          range = sheet.Range[ iRow, iColumnIndex + 1 ];
          style = table.Appearance.SummaryFieldCell;
          CopyStyle( style, range );
        }

        range.Text = column.GetDisplayText( summary.ParentGroup );
      }

      return 1;
    }
    /// <summary>
    /// Performs all necessary operations (grouping necessary rows, etc.)
    /// to set current group level to the desired level.
    /// </summary>
    /// <param name="stackStartIndexes">Stack with starting index of previously found groups.</param>
    /// <param name="iCurLevel">Current group level.</param>
    /// <param name="iGroupLevel">Group level to set.</param>
    /// <param name="sheet">Destination worksheet.</param>
    /// <param name="index">Current row zero-based index.</param>
    /// <param name="bStartNewGroup">Indicates whether new group should be started at index.</param>
    private void AdjustGroupingLevel( Stack stackStartIndexes, ref int iCurLevel, int iGroupLevel,
      IWorksheet sheet, int index, bool bStartNewGroup )
    {
      if( stackStartIndexes == null )
        throw new ArgumentNullException( "stackStartIndexes" );

      if( sheet == null )
        throw new ArgumentNullException( "sheet" );

      if( iCurLevel < 0 || iCurLevel < iGroupLevel )
      {
        iCurLevel = iGroupLevel;
        stackStartIndexes.Push( index );
      }
      else if( iCurLevel >= iGroupLevel )
      {
        int iLastLevel = bStartNewGroup ? iGroupLevel : iGroupLevel + 1;

        for( ; iCurLevel >= iLastLevel; iCurLevel-- )
        {
          int iStartIndex = ( int )stackStartIndexes.Pop() + 1;
          int iEndIndex = index;

          if( iStartIndex != iEndIndex )
          {
            sheet.Range[ iStartIndex + 1, 1, index, 1 ].Group( ExcelGroupBy.ByRows, false );
          }
        }

        if( bStartNewGroup )
        {
          iCurLevel++;
          stackStartIndexes.Push( index );
        }
      }
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="row"></param>
    /// <param name="sheet"></param>
    /// <param name="index"></param>
    /// <param name="options"></param>
    /// <param name="iIndexInGroup"></param>
    /// <returns></returns>
    private int ExportRecordRow( RecordRow row, IWorksheet sheet, int index, ConverterOptions options
      , int iIndexInGroup )
    {
      if( row == null )
        throw new ArgumentNullException( "row" );

      if( sheet == null )
        throw new ArgumentNullException( "sheet" );

      if( index < 0 )
        throw new ArgumentOutOfRangeException( "index", index, "Value can not be less 0" );

      if( iIndexInGroup < 0 )
        throw new ArgumentOutOfRangeException( "iIndexInGroup", iIndexInGroup, "Value can not be less than 0" );

      FieldDescriptorCollection arrFields = row.ParentTable.TableDescriptor.Fields;
      int iRowIndex = index + 1;

      GridTableDescriptor table = ( GridTableDescriptor )row.ParentTableDescriptor;
      GridColumnDescriptorCollection arrColumns = table.Columns;

      for( int i = 0, len = arrColumns.Count; i < len; i++ )
      {
        GridColumnDescriptor column = arrColumns[ i ];
        FieldDescriptor field = arrFields[ column.Name ];

        IRange range = sheet.Range[ iRowIndex, i + 1 ];
        object value = row.ParentRecord.GetValue( field );
        range.Text = ( value == null ) ? string.Empty : value.ToString();
        
        if( options == ConverterOptions.ConvertVisible )
        {
          ExportStyle( row, field, range, options );
        }
        else
        {
          if( iIndexInGroup % 2 == 0 )
          {
            CopyStyle( table.Appearance.RecordFieldCell, range );
          }
          else
          {
            CopyStyle( table.Appearance.AlternateRecordFieldCell, range );
          }
        }
      }

      return 1;
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="sourceElement"></param>
    /// <param name="sourceField"></param>
    /// <param name="destRange"></param>
    /// <param name="options"></param>
    private void ExportStyle( Element sourceElement, FieldDescriptor sourceField, IRange destRange,
      ConverterOptions options )
    {
      if( destRange == null )
        throw new ArgumentNullException( "destRange" );

      if( sourceElement == null )
        throw new ArgumentNullException( "sourceElement" );

//      if( sourceField == null )
//        throw new ArgumentNullException( "sourceField" );

      GridTable table = sourceElement.ParentTable as GridTable;

      if( table == null )
      {
        Debug.WriteLine( "Bad table type" );
        return;
      }

      GridTableCellStyleInfo style;

      //if( options == ConverterOptions.ConvertVisible )
      {
        string strFieldName = sourceField != null ? sourceField.Name : null;
        style = table.GetTableCellStyle( sourceElement, strFieldName );
      }
//      else // if( options == ConverterOptions.ConvertAll )
//      {
//        style = table.Appearance.AnyRecordFieldCell;
//      }

      CopyStyle( style, destRange );
    }
    /// <summary>
    /// Exports caption row.
    /// </summary>
    /// <param name="captionRow">Source caption row.</param>
    /// <param name="sheet">Destionation worksheet.</param>
    /// <param name="index">Capiton's zero-based row index in the worksheet.</param>
    /// <param name="options">Convert options.</param>
    /// <param name="iCurrentGroupLevel">Current group level.</param>
    /// <returns>Zero-based row index that point to the row after caption.</returns>
    private int ExportCaption( GridCaptionRow captionRow, IWorksheet sheet, int index,
      ConverterOptions options, ref int iCurrentGroupLevel, Stack stackStartIndexes )
    {
      if( captionRow == null )
        throw new ArgumentNullException( "captionRow" );

      if( sheet == null )
        throw new ArgumentNullException( "sheet" );

      if( index < 0 )
        throw new ArgumentOutOfRangeException( "index", index, "Value can not be less 0" );

      IRange range = sheet.Range[ index + 1, 1 ];
      range.Text = GetCaptionText( captionRow );

      CopyStyle( captionRow.Appearance.GroupCaptionCell, range );
      GridTableModel tableModel = captionRow.ParentTable.TableModel;
      int iRealColCount = tableModel.ColCount - tableModel.GetColumnIndentCount();

      sheet.Range[ index + 1, 1, index + 1, iRealColCount ].Merge();

      int iGroupLevel = captionRow.GroupLevel;
      AdjustGroupingLevel( stackStartIndexes, ref iCurrentGroupLevel, iGroupLevel, sheet, index, true );
      
      return 1;
    }
    /// <summary>
    /// Exports group to the excel worksheet.
    /// </summary>
    /// <param name="group">Group to export.</param>
    /// <param name="sheet">Destination worksheet.</param>
    /// <param name="index">Zero-based row index.</Bparam>
    /// <returns>Zero-based row index that point to the row after group rows.</returns>
    private int ExportGroup( GridGroup group, IWorksheet sheet, int index )
    {
      if( group == null )
        throw new ArgumentNullException( "group" );

      if( sheet == null )
        throw new ArgumentNullException( "sheet" );

      if( index < 0 )
        throw new ArgumentOutOfRangeException( "index", index, "Value can not be less 0" );

      RecordsInDetailsCollection arrRecords = group.Records;
      return ExportElements( arrRecords, sheet, index, ConverterOptions.ConvertAll );
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="table"></param>
    /// <param name="sheet"></param>
    /// <param name="index"></param>
    /// <param name="options"></param>
    /// <returns></returns>
    private int ExportNestedTable( NestedTable table, IWorksheet sheet, int index, ConverterOptions options )
    {
      IList elements = null;
      ChildTable childTable = table.ChildTable;

      if( options == ConverterOptions.ConvertVisible )
      {
        elements = childTable.DisplayElements;
      }
      else
      {
        elements = childTable.Elements;
      }

      return ExportElements( elements, sheet, index, options );
    }
    /// <summary>
    /// Copies column width settings from grid model into excel worksheet.
    /// </summary>
    /// <param name="gridModel">Source grid.</param>
    /// <param name="sheet">Destination worksheet.</param>
    private void CopyColumnWidthFromGrid( GridTableModel gridModel, IWorksheet sheet )
    {
      if( gridModel == null )
        throw new ArgumentNullException( "gridModel" );

      if( sheet == null )
        throw new ArgumentNullException( "sheet" );

      GridTableDescriptor table = gridModel.Table.TableDescriptor;
      GridColumnDescriptorCollection arrColumns = table.Columns;

      for( int i = 0, len = arrColumns.Count; i < len; i++ )
      {
        GridColumnDescriptor column = arrColumns[ i ];
        int iWidthInPixels = column.Width;
        double dWidth = sheet.PixelsToColumnWidth( iWidthInPixels );
        sheet.SetColumnWidth( i + 1, dWidth );
      }
    }
    #endregion

    #region Class static methods
    /// <summary>
    /// 
    /// </summary>
    /// <param name="caption"></param>
    /// <returns></returns>
    public static string GetCaptionText( GridCaptionRow caption )
    {
      return GetGroupCaptionText( caption.ParentGroup );
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="group"></param>
    /// <returns></returns>
    public static string GetGroupCaptionText( Group group )
    {
      IGridGroupOptionsSource g = group as IGridGroupOptionsSource;
      string captionText;

      if (g != null)
      {
        captionText = g.GroupOptions.CaptionText;
      }
      else
      {
        captionText = "{CategoryCaption}: {Category} - {RecordCount} Items";
      }
 
      return GetGroupCaptionDisplayText(group, captionText);
    }

    /// <summary>
    /// 
    /// </summary>
    /// <param name="group"></param>
    /// <param name="format"></param>
    /// <returns></returns>
    public static string GetGroupCaptionDisplayText( Group group, string format )
    {
      GridTableDescriptor tableDescriptor = ( GridTableDescriptor )group.ParentTableDescriptor;

      bool raiseException = false;
      ArrayList al = new ArrayList();
      int iOpenBracketPos = format.IndexOf( "{" );
      StringBuilder sb = new StringBuilder();
      int iCloseBracketPos = 0;

      if( iOpenBracketPos == -1 )
      {
        sb.Append( format );
      }
      else
      {
        sb.Append( format.Substring( 0, iOpenBracketPos + 1 ) );
      }
 
      while( iOpenBracketPos != -1 )
      {
        iCloseBracketPos = format.IndexOf( "}", iOpenBracketPos );

        if( iOpenBracketPos != -1 && iCloseBracketPos != -1 && iCloseBracketPos > iOpenBracketPos )
        {
          int n3 = format.IndexOfAny( new char[] { '}', ':' }, iOpenBracketPos );
          string name = format.Substring( iOpenBracketPos + 1, n3 - iOpenBracketPos - 1 );

          sb.Append( al.Count.ToString() );
          object obj = "";

          if( name == "TableName" )
          {
            obj = group.ParentTableDescriptor.Name;
          }
          else if( name == "CategoryName" )
          {
            obj = group.Name;
          }
          else if( name == "CategoryCaption" )
          {
            string fieldName = group.Name;

            GridColumnDescriptor cd = group.GroupLevel >= 0
              ? tableDescriptor.Columns.FindByMappingName( fieldName )
              : null;

            obj = ( cd != null ) ? cd.HeaderText : fieldName;
          }
          else if( name == "Category" )
          {
            obj = group.Category;
          }
          else if( name == "RecordCount" )
          {
            obj = group.GetFilteredRecordCount();
          }
          else
          {
            GridTableDescriptor td = ( ( GridTableDescriptor )group.ParentTableDescriptor );
            int summaryRowNum;
            int dot = name.IndexOf( '.' );

            if( dot != -1 )
            {
              string rowName = name.Substring( 0, dot );
              name = name.Substring( dot + 1 );
              summaryRowNum = td.SummaryRows.IndexOf(rowName);
            }
            else
            {
              summaryRowNum = td.SummaryRows.IndexOf( "GroupCaption" );
            }

            if( summaryRowNum != -1 )
            {
              GridSummaryColumnDescriptor scd = td.SummaryRows[ summaryRowNum ].SummaryColumns[ name ];

              if( scd != null )
              {
                Group g = group;
                Syncfusion.Collections.BinaryTree.ITreeTableSummary[] sums =
                  g.GetSummaries( group.ParentTable );
                int ndx = scd.GetSummaryIndex();
                obj = ( scd.GetDisplayText( sums[ ndx ] ) );
              }
            }
          }
          al.Add( obj );
          iOpenBracketPos = format.IndexOf( "{", iCloseBracketPos );

          if( iOpenBracketPos == -1 )
          {
            sb.Append( format.Substring( n3 ) );
          }
          else
          {
            sb.Append( format.Substring( n3, iOpenBracketPos - n3 + 1 ) );
          }
        }
        else
        {
          if( raiseException )
          {
            throw new FormatException( "No closing char found: " + sb.ToString() );
          }
          else
          {
            break;
          }
        }
      }
      string formatString = sb.ToString();

      try
      {
        return String.Format(formatString, al.ToArray());
      }
      catch (Exception ex)
      {
        return formatString + ": " + ex.Message;
      }
    }

    #endregion

    #region Class Import into Grid from Excel
    /// <summary>
    /// Converts excel file into GridControl.
    /// </summary>
    /// <param name="strFileName">File to convert.</param>
    /// <param name="grid">Grid which will get converted data.</param>
    public void ExcelToGrid( string strFileName, GridModel grid )
    {
      if( strFileName == null )
        throw new ArgumentNullException( "strFileName" );

      if( strFileName.Length == 0 )
        throw new ArgumentException( "strFileName - string can not be empty" );

      if( grid == null )
        throw new ArgumentNullException( "grid" );

      ExcelEngine engine = new ExcelEngine();
      IWorkbook book = engine.Excel.Workbooks.Open( strFileName );
      IWorksheet sheet = book.Worksheets[ 0 ];
      ExcelToGrid( sheet, grid );

      book.Close( false );
    }
    /// <summary>
    /// Exports data from excel worksheet into grid.
    /// </summary>
    /// <param name="sheet">Worksheet to export from.</param>
    /// <param name="grid">Grid to export into.</param>
    public void ExcelToGrid( IWorksheet sheet, GridModel grid )
    {
      if( sheet == null )
        throw new ArgumentNullException( "sheet" );

      if( grid == null )
        throw new ArgumentNullException( "grid" );

      IRange range = sheet.UsedRange;
      int iFirstRow = range.Row;
      int iLastRow  = range.End.Row;
      int iFirstCol = range.Column;
      int iLastCol  = range.End.Column;

      grid.BeginUpdate();
      grid.RowCount = iLastRow + 1;
      grid.ColCount = iLastCol + 1;

      for( int iRow = iFirstRow; iRow <= iLastRow; iRow++ )
      {
        for( int iCol = iFirstCol; iCol <= iLastCol; iCol++ )
        {
          IRange rangeToConvert = sheet.Range[ iRow, iCol ];
          ConvertExcelRangeToGrid( rangeToConvert, grid );
        }
      }

      CopyRowHeightToGrid( sheet, grid );
      CopyColumnWidthToGrid( sheet, grid );
      CopyMergesToGrid( sheet, grid );
      grid.EndUpdate();
    }
    /// <summary>
    /// Converts excel workbook into array of grid models.
    /// </summary>
    /// <param name="strFileName">File to convert.</param>
    /// <param name="arrIndexes">Worksheet's indexes to convert.</param>
    /// <param name="arrModels">Array of models to convert into.</param>
    public void ExcelToGrid( string strFileName, int[] arrIndexes, GridModel[] arrModels )
    {
      if( strFileName == null )
        throw new ArgumentNullException( "strFileName" );

      if( strFileName.Length == 0 )
        throw new ArgumentException( "strFileName - string can not be empty" );

      if( arrIndexes == null )
        throw new ArgumentNullException( "arrIndexes" );

      if( arrModels == null )
        throw new ArgumentNullException( "arrModels" );

      if( arrIndexes.Length != arrModels.Length )
        throw new ArgumentException( "Indexes and models do not correspond each other" );

      ExcelEngine engine = new ExcelEngine();
      IWorkbook book = engine.Excel.Workbooks.Open( strFileName );
      int iSheetsCount = book.Worksheets.Count;

      for( int i = 0, len = arrIndexes.Length; i < len; i++ )
      {
        int index = arrIndexes[ i ];
        if( index >= iSheetsCount ) continue;

        IWorksheet sheet = book.Worksheets[ index ];
        GridModel model = arrModels[ i ];
        ExcelToGrid( sheet, model );
      }

      book.Close();
    }
    /// <summary>
    /// Converts excel workbook into array of grid models.
    /// </summary>
    /// <param name="strFileName">File to convert.</param>
    /// <param name="arrRanges">Array of worksheet ranges to convert.</param>
    /// <param name="arrModels">Array of GridModels to convert into.</param>
    public void ExcelToGrid( string strFileName, IndexRange[] arrRanges, GridModel[] arrModels )
    {
      if( strFileName == null )
        throw new ArgumentNullException( "strFileName" );

      if( strFileName.Length == 0 )
        throw new ArgumentException( "strFileName - string can not be empty" );

      if( arrRanges == null )
        throw new ArgumentNullException( "arrRanges" );

      if( arrModels == null )
        throw new ArgumentNullException( "arrModels" );

      ExcelEngine engine = new ExcelEngine();
      IWorkbook book = engine.Excel.Workbooks.Open( strFileName );
      int iSheetsCount = book.Worksheets.Count;
      int iModelIndex = 0;
      int iModelCount = arrModels.Length;

      for( int i = 0, len = arrRanges.Length; i < len; i++ )
      {
        IndexRange range = arrRanges[ i ];
        for( int j = range.FirstIndex, last = range.LastIndex; j <= last; j++ )
        {
          if( j >= iSheetsCount ) break;

          IWorksheet sheet = book.Worksheets[ j ];

          if( iModelIndex >= iModelCount )
          {
            throw new ArgumentOutOfRangeException( "Model index was out of range" );
          }

          ExcelToGrid( sheet, arrModels[ iModelIndex ] );
          iModelIndex++;
        }
      }

      book.Close();
    }
    /// <summary>
    /// Converts excel file into array of grid models.
    /// </summary>
    /// <param name="strFileName">File name to convert.</param>
    /// <returns>Corresponding grid models.</returns>
    public GridModel[] ExcelToGrid( string strFileName )
    {
      if( strFileName == null )
        throw new ArgumentNullException( "strFileName" );

      if( strFileName.Length == 0 )
        throw new ArgumentException( "strFileName - string can not be empty" );

      ExcelEngine engine = new ExcelEngine();
      IWorkbook book = engine.Excel.Workbooks.Open( strFileName );
      return ExcelToGrid( book );
    }
    /// <summary>
    /// Converts excel workbook into array of grid models.
    /// </summary>
    /// <param name="book">Source workbook to convert.</param>
    /// <returns>Corresponding grid models.</returns>
    public GridModel[] ExcelToGrid( IWorkbook book )
    {
      if( book == null )
        throw new ArgumentNullException( "book" );

      int iSheetsCount = book.Worksheets.Count;

      GridModel[] arrResult = new GridModel[ iSheetsCount ];

      for( int i = 0; i < iSheetsCount; i++ )
      {
        GridModel model = new GridModel();
        ExcelToGrid( book.Worksheets[ i ], model );
        arrResult[ i ] = model;
      }

      book.Close( false );

      return arrResult;
    }
    /// <summary>
    /// Puts data from IRange into GridControl.
    /// </summary>
    /// <param name="rangeToConvert">Range to convert.</param>
    /// <param name="grid">Grid to convert into.</param>
    private void ConvertExcelRangeToGrid( IRange rangeToConvert, GridModel grid )
    {
      if( rangeToConvert == null )
        throw new ArgumentNullException( "rangeToConvert" );

      if( grid == null )
        throw new ArgumentNullException( "grid" );

      if( !rangeToConvert.IsInitialized ) return;

      GridStyleInfo cell = grid[ rangeToConvert.Row, rangeToConvert.Column ];
      cell.Text = rangeToConvert.Value;
      
      if( rangeToConvert.HasFormula || rangeToConvert.HasFormulaArray )
      {
        cell.CellType = "FormulaCell";
      }

      SetCellStyle( rangeToConvert.CellStyle, cell );
    }
    /// <summary>
    /// Sets grid cell style based on excel style.
    /// </summary>
    /// <param name="excelStyle">Source style.</param>
    /// <param name="cell">Destination cell.</param>
    private void SetCellStyle( IStyle excelStyle, GridStyleInfo cell )
    {
      if( excelStyle == null )
        throw new ArgumentNullException( "excelStyle" );

      if( cell == null )
        throw new ArgumentNullException( "cell" );

      if( !excelStyle.IsInitialized ) return;

      SetBorders( excelStyle, cell );
      SetAlignment( excelStyle, cell );
      SetFont( excelStyle, cell );
      SetBrush( excelStyle, cell );
    }
    /// <summary>
    /// Sets grid cell borders based on excel style.
    /// </summary>
    /// <param name="excelStyle">Source style.</param>
    /// <param name="cell">Destination cell.</param>
    private void SetBorders( IStyle excelStyle, GridStyleInfo cell )
    {
      if( excelStyle == null )
        throw new ArgumentNullException( "excelStyle" );

      if( cell == null )
        throw new ArgumentNullException( "cell" );

      GridBordersInfo borders = cell.Borders;//new GridBordersInfo();
      IBorders rangeBorders = excelStyle.Borders;

      borders.Bottom = ConvertExcelBorder( rangeBorders[ ExcelBordersIndex.EdgeBottom ] );
      borders.Top = ConvertExcelBorder( rangeBorders[ ExcelBordersIndex.EdgeTop ] );
      borders.Left = ConvertExcelBorder( rangeBorders[ ExcelBordersIndex.EdgeLeft ] );
      borders.Right = ConvertExcelBorder( rangeBorders[ ExcelBordersIndex.EdgeRight ] );
    }
    /// <summary>
    /// Converts excel border into grid border.
    /// </summary>
    /// <param name="border">Border to convert.</param>
    private GridBorder ConvertExcelBorder( IBorder border )
    {
      if( border == null )
        throw new ArgumentNullException( "border" );

      GridBorderStyle borderStyle = GetGridBorderStyle( border.LineStyle );
      GridBorderWeight borderWeight = GetGridBorderWeight( border.LineStyle );

      return new GridBorder( borderStyle, border.ColorRGB, borderWeight );
    }
    /// <summary>
    /// Converts ExcelLineStyle into GridBorderStyle.
    /// </summary>
    /// <param name="excelLineStyle">Line style to convert.</param>
    /// <returns>Converted value.</returns>
    private GridBorderStyle GetGridBorderStyle( ExcelLineStyle excelLineStyle )
    {
      switch( excelLineStyle )
      {
        case ExcelLineStyle.None:
          return GridBorderStyle.None;

        case ExcelLineStyle.Thin:
        case ExcelLineStyle.Medium:
        case ExcelLineStyle.Thick:
        case ExcelLineStyle.Hair:
        case ExcelLineStyle.Double:
          return GridBorderStyle.Solid;

        case ExcelLineStyle.Dashed:
        case ExcelLineStyle.Medium_dashed:
          return GridBorderStyle.Dashed;

        case ExcelLineStyle.Dotted:
          return GridBorderStyle.Dotted;


        case ExcelLineStyle.Dash_dot:
        case ExcelLineStyle.Medium_dash_dot:
        case ExcelLineStyle.Slanted_dash_dot:
          return GridBorderStyle.DashDot;

        case ExcelLineStyle.Dash_dot_dot:
        case ExcelLineStyle.Medium_dash_dot_dot:
          return GridBorderStyle.DashDotDot;

        default:
          throw new ArgumentOutOfRangeException( "excelLineStyle" );
      }
    }
    /// <summary>
    /// Converts ExcelLineStyle into GridBorderWeight.
    /// </summary>
    /// <param name="excelLineStyle">Line style to convert.</param>
    /// <returns>Covnerted value.</returns>
    private GridBorderWeight GetGridBorderWeight( ExcelLineStyle excelLineStyle )
    {
      switch( excelLineStyle )
      {
        case ExcelLineStyle.None:
        case ExcelLineStyle.Hair:
          return GridBorderWeight.ExtraThin;


        case ExcelLineStyle.Thin:
        case ExcelLineStyle.Dashed:
        case ExcelLineStyle.Dotted:
        case ExcelLineStyle.Dash_dot:
        case ExcelLineStyle.Slanted_dash_dot:
        case ExcelLineStyle.Dash_dot_dot:
          return GridBorderWeight.Thin;

        case ExcelLineStyle.Medium:
        case ExcelLineStyle.Medium_dashed:
        case ExcelLineStyle.Medium_dash_dot:
        case ExcelLineStyle.Medium_dash_dot_dot:
          return GridBorderWeight.Medium;

        case ExcelLineStyle.Thick:
        case ExcelLineStyle.Double:
          return GridBorderWeight.Thick;

        default:
          throw new ArgumentOutOfRangeException( "excelLineStyle" );
      }
    }
    /// <summary>
    /// Sets alignment of the GridStyleInfo.
    /// </summary>
    /// <param name="excelStyle">Excel style to get alignment block from.</param>
    /// <param name="cell">GridStyleInfo to set alignment block in.</param>
    private void SetAlignment( IStyle excelStyle, GridStyleInfo cell )
    {
      if( excelStyle == null )
        throw new ArgumentNullException( "excelStyle" );

      if( cell == null )
        throw new ArgumentNullException( "cell" );

      cell.HorizontalAlignment = GetGridHAlign( excelStyle.HorizontalAlignment );
      cell.VerticalAlignment = GetGridVAlign( excelStyle.VerticalAlignment );
      cell.WrapText = excelStyle.WrapText;
    }
    /// <summary>
    /// Converts ExcelHAling into GridHorizontalAlignment.
    /// </summary>
    /// <param name="align">Alignment to convert.</param>
    /// <returns>Covnerted value.</returns>
    private GridHorizontalAlignment GetGridHAlign( ExcelHAlign align )
    {
      switch( align )
      {
        case ExcelHAlign.HAlignLeft: return GridHorizontalAlignment.Left;
        case ExcelHAlign.HAlignRight: return GridHorizontalAlignment.Right;

        default: return GridHorizontalAlignment.Center;
      }
    }
    /// <summary>
    /// Converts ExcelVAling into GridVerticalAlignment.
    /// </summary>
    /// <param name="align">Alignment to convert.</param>
    /// <returns>Converted value.</returns>
    private GridVerticalAlignment GetGridVAlign( ExcelVAlign align )
    {
      switch( align )
      {
        case ExcelVAlign.VAlignBottom: return GridVerticalAlignment.Bottom;
        case ExcelVAlign.VAlignCenter: return GridVerticalAlignment.Middle;
        case ExcelVAlign.VAlignTop:    return GridVerticalAlignment.Top;

        default: return GridVerticalAlignment.Middle;
      }
    }
    /// <summary>
    /// Sets font for the cell based on ExcelRW style.
    /// </summary>
    /// <param name="excelStyle">ExcelRW style to get font setting from.</param>
    /// <param name="cell">Cell style to set font.</param>
    private void SetFont( IStyle excelStyle, GridStyleInfo cell )
    {
      if( excelStyle == null )
        throw new ArgumentNullException( "excelStyle" );

      if( cell == null )
        throw new ArgumentNullException( "cell" );

      IFont excelFont = excelStyle.Font;

      GridFontInfo gridFont = cell.Font;
      gridFont.Bold = excelFont.Bold;
      gridFont.Italic = excelFont.Italic;
      gridFont.Underline = ( excelFont.Underline != ExcelUnderline.None );
      gridFont.Facename = excelFont.FontName;
      gridFont.Orientation = excelStyle.Rotation;
      gridFont.Size = excelFont.Size;
      gridFont.Strikeout = excelFont.Strikethrough;
      gridFont.Unit = GraphicsUnit.Point;

      cell.TextColor = excelFont.RGBColor;
    }
    /// <summary>
    /// Sets cell's brush.
    /// </summary>
    /// <param name="excelStyle">Excel style containing information about brush.</param>
    /// <param name="cell">Style info to set brush in.</param>
    private void SetBrush( IStyle excelStyle, GridStyleInfo cell )
    {
      if( excelStyle == null )
        throw new ArgumentNullException( "excelStyle" );

      if( cell == null )
        throw new ArgumentNullException( "cell" );

      PatternStyle patternStyle = GetClosestGridPattern( excelStyle.FillPattern );

      //if( patternStyle == PatternStyle.None ) return;
      BrushInfo brush;

      if( excelStyle.FillPattern == ExcelPattern.Solid )
      {
        brush = new BrushInfo( excelStyle.FillBackgroundRGB );
      }
      else
      {
        brush = new BrushInfo( patternStyle, excelStyle.FillBackgroundRGB,
          excelStyle.FillForegroundRGB );
      }

      cell.Interior = brush;
    }
    /// <summary>
    /// Converts ExcelPattern into PatternStyle.
    /// </summary>
    /// <param name="pattern">ExcelPattern to convert.</param>
    /// <returns>Converted value.</returns>
    private PatternStyle GetClosestGridPattern( ExcelPattern pattern )
    {
      switch( pattern )
      {
        case ExcelPattern.Angle:                return PatternStyle.Cross;
        case ExcelPattern.DarkDownwardDiagonal: return PatternStyle.DarkDownwardDiagonal;
        case ExcelPattern.DarkHorizontal:       return PatternStyle.DarkHorizontal;
        case ExcelPattern.DarkUpwardDiagonal:   return PatternStyle.DarkUpwardDiagonal;
        case ExcelPattern.DarkVertical:         return PatternStyle.DarkVertical;

        case ExcelPattern.ForwardDiagonal:      return PatternStyle.ForwardDiagonal;
        case ExcelPattern.Horizontal:           return PatternStyle.Horizontal;
        case ExcelPattern.LightDownwardDiagonal:return PatternStyle.LightDownwardDiagonal;
        case ExcelPattern.LightUpwardDiagonal:  return PatternStyle.LightUpwardDiagonal;
        case ExcelPattern.Vertical:             return PatternStyle.Vertical;
        case ExcelPattern.None:                 return PatternStyle.None;
        case ExcelPattern.Percent05:            return PatternStyle.Percent05;
        case ExcelPattern.Percent10:            return PatternStyle.Percent10;
        case ExcelPattern.Percent25:            return PatternStyle.Percent25;

        case ExcelPattern.Percent50:            return PatternStyle.Percent50;

        case ExcelPattern.Percent60:            return PatternStyle.Percent60;
        case ExcelPattern.Percent70:            return PatternStyle.Percent70;
        case ExcelPattern.Percent75:            return PatternStyle.Percent75;
        case ExcelPattern.Solid:                return PatternStyle.ZigZag;

        default: throw new ArgumentOutOfRangeException( "pattern" );
      }
    }
    /// <summary>
    /// Copies row height information from excel worksheet into grid.
    /// </summary>
    /// <param name="sheet">Sheet to copy from.</param>
    /// <param name="grid">Grid to copy into.</param>
    private void CopyRowHeightToGrid( IWorksheet sheet, GridModel grid )
    {
      if( grid == null )
        throw new ArgumentNullException( "grid" );

      if( sheet == null )
        throw new ArgumentNullException( "sheet" );

      IRange usedRange = sheet.UsedRange;

      for( int i = usedRange.Row, last = usedRange.End.Row; i <= last; i++ )
      {
        grid.RowHeights[ i ] = sheet.GetRowHeightInPixels( i );
      }
    }
    /// <summary>
    /// Copies column width information from excel worksheet into grid.
    /// </summary>
    /// <param name="sheet">Sheet to copy from.</param>
    /// <param name="grid">Grid to copy into.</param>
    private void CopyColumnWidthToGrid( IWorksheet sheet, GridModel grid )
    {
      if( grid == null )
        throw new ArgumentNullException( "grid" );

      if( sheet == null )
        throw new ArgumentNullException( "sheet" );

      IRange usedRange = sheet.UsedRange;

      for( int i = usedRange.Column, last = usedRange.End.Column; i < last; i++ )
      {
        grid.Model.ColWidths[ i ] = sheet.GetColumnWidthInPixels( i );
      }
    }
    /// <summary>
    /// Copies merged ranges into grid.
    /// </summary>
    /// <param name="sheet">Destination worksheet.</param>
    /// <param name="grid">Source grid.</param>
    private void CopyMergesToGrid( IWorksheet sheet, GridModel grid )
    {
      if( grid == null )
        throw new ArgumentNullException( "grid" );

      if( sheet == null )
        throw new ArgumentNullException( "sheet" );

      GridModelCoveredRanges coveredRanges = grid.CoveredRanges;
      IRange[] arrMerges = sheet.MergedCells;

      for( int i = 0, len = arrMerges.Length; i < len; i++ )
      {
        IRange curRange = arrMerges[ i ];
        GridRangeInfo mergedRange = GridRangeInfo.Cells( curRange.Row, curRange.Column,
          curRange.End.Row, curRange.End.Column );

        coveredRanges.Add( mergedRange );
      }
    }
    #endregion
  }

  /// <summary>
  /// 
  /// </summary>
  public enum ConverterOptions
  {
    ConvertVisible,
    ConvertAll,
  }
}
